/*
 * Javolution - Java(TM) Solution for Real-Time and Embedded Systems
 * Copyright (C) 2005 - Javolution (http://javolution.org/)
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software is
 * freely granted, provided that this notice is preserved.
 */
package javolution.xml;

import j2me.lang.CharSequence;
import j2me.io.Serializable;
import j2me.lang.UnsupportedOperationException;
import javolution.lang.TypeFormat;
import javolution.realtime.ArrayPool;
import javolution.realtime.RealtimeObject;

/**
 * <p> This class represents the text that is not markup and constitutes
 *     the "Character Data" of a XML document.</p>
 *     
 * <p> During deserialization, instances of this class are generated for 
 *     character data containing at least one non-whitespace character.</p>
 *     
 * <p> During serialization, instances of this class are written in a 
 *     "CDATA" section (<code>&lt;![CDATA[...]]&gt;</code>).</p>
 *
 * @author  <a href="mailto:jean-marie@dautelle.com">Jean-Marie Dautelle</a>
 * @version 2.2, January 8, 2005
 */
public final class CharacterData extends RealtimeObject implements CharSequence, Serializable {

    /**
     * Holds the object factory.
     */
    private static final Factory FACTORY = new Factory() {
        public Object create() {
            return new CharacterData();
        }
    };
    
    /**
     * Holds the characters.
     */
    char[] _chars;

    /**
     * Holds the number of characters.
     */
    int _length;

    /**
     * Returns the character data for the specified string.
     *
     * @param  str the <code>String</code> source.
     * @return the corresponding character data instance.
     */
    public static CharacterData valueOf(String str) {
        CharacterData cd = (CharacterData) FACTORY.object();
        int length = str.length();
        cd._chars = (char[]) ArrayPool.charArray(length).next();
        for (int i=0; i < length;) {
            cd._chars[i] = str.charAt(i++);
        }
        cd._length = length;
        return cd;
    }

    /**
     * Returns the character data for the specified character sequence.
     *
     * @param  csq the <code>CharSequence</code> source.
     * @return the corresponding character data instance.
     */
    public static CharacterData valueOf(CharSequence csq) {
        CharacterData cd = (CharacterData) FACTORY.object();
        int length = csq.length();
        cd._chars = (char[]) ArrayPool.charArray(length).next();
        for (int i=0; i < length;) {
            cd._chars[i] = csq.charAt(i++);
        }
        cd._length = length;
        return cd;
    }

    /**
     * Returns the character data that contains the characters from the specified 
     * subarray of characters.
     *
     * @param data the source of the characters.
     * @param offset the index of the first character in the data soure.
     * @param length the length of the subarray.
     * @return the corresponding character data instance.
     */
    public static CharacterData valueOf(char[] data, int offset, int length) {
        CharacterData cd = (CharacterData) FACTORY.object();
        cd._chars = (char[]) ArrayPool.charArray(length).next();
        for (int i=0; i < length;) {
            cd._chars[i] = data[offset + i++];
        }
        cd._length = length;
        return cd;
    }

    /**
     * Default constructor.
     */
    private CharacterData() {
    }

    /**
     * Returns the length of this character data.
     *
     * @return the number of characters (16-bits Unicode).
     */
    public int length() {
        return _length;
    }
  
    /**
     * Returns the character at the specified index.
     *
     * @param  index the index of the character.
     * @return the character at the specified index.
     * @throws IndexOutOfBoundsException if index is negative, or index 
     *         is equal or greater than <code>this.length()</code>.
     */
    public char charAt(int index) {
        if ((index < 0) || (index >= _length)) throw new IndexOutOfBoundsException();
        return _chars[index];
    }

    /**
     * Returns a new character data that is a subsequence of this one.
     *
     * @param  start the index of the first character inclusive.
     * @param  end the index of the last character exclusive.
     * @return the corresponding character data instance.
     * @throws IndexOutOfBoundsException if start or end are negative, start 
     *         is greater than end, or end is greater than 
     *         <code>this.length()</code>.
     */
    public CharSequence subSequence(int start, int end) {
        return valueOf(_chars, start, end - start);
    }

    /**
     * Compares this character data against the specified object. This method 
     * uses a {@link TypeFormat#LEXICAL_COMPARATOR lexical comparator}
     * to make this determination.
     * 
     * @param  that the object to compare with.
     * @return <code>TypeFormat.LEXICAL_COMPARATOR.compare(this, that) == 0</code>
     */
    public final boolean equals(Object that) {
        if ((that instanceof CharSequence) || (that instanceof String)) {
            return TypeFormat.LEXICAL_COMPARATOR.compare(this, that) == 0;
        }
        return false;
    }

    /**
     * Returns the hash code for this character data.
     *
     * @return the hash code value.
     */
    public final int hashCode() {
        int h = 0;
        for (int i = 0; i < _length;) {
            h = 31 * h + _chars[i++];
        }
        return h;
    }

    // Overrides.
    public void move(ContextSpace cs) {
        throw new UnsupportedOperationException();
    }
}